/*
 * DifferentialDriveNavigator.java
 * 
 * Copyright 2004-2009 by RidgeSoft, LLC., PO Box 482, Pleasanton, CA  94566, U.S.A.
 * www.ridgesoft.com
 *
 * RidgeSoft grants you the right to use, modify, make derivative works and
 * redistribute this source file provided you do not remove this copyright notice.
 */

import com.ridgesoft.robotics.Motor;
import com.ridgesoft.robotics.Localizer;
import com.ridgesoft.robotics.Pose;

public class DifferentialDriveNavigator extends Thread implements Navigator {
    private static final float PI = 3.14159f;
    private static final float TWO_PI = PI * 2.0f;

    private static final int STOP = 0;
    private static final int GO = 1;
    private static final int MOVE_TO = 2;
    private static final int ROTATE = 3;

    private int state;
    private int period;
    private Motor leftMotor;
    private Motor rightMotor;

    private Localizer localizer;
    private int drivePower;
    private int rotatePower;
    private float destinationX;
    private float destinationY;
    private float goToThreshold;
    private float targetHeading;
    private float gain;
    private float rotateThreshold;

    public DifferentialDriveNavigator(Motor leftMotor, Motor rightMotor,
            Localizer localizer, int drivePower, int rotatePower, float gain,
            float goToThreshold, float rotateThreshold, int threadPriority,
            int period) {
        this.leftMotor = leftMotor;
        this.rightMotor = rightMotor;
        this.localizer = localizer;
        this.drivePower = drivePower;
        this.gain = gain;
        this.goToThreshold = goToThreshold;
        this.rotatePower = rotatePower;
        this.rotateThreshold = rotateThreshold;
        this.period = period;
        state = STOP;
        setPriority(threadPriority);
        setDaemon(true);
        start();
    }

    public synchronized void moveTo(float x, float y, boolean wait) {
        destinationX = x;
        destinationY = y;
        state = MOVE_TO;
        if (wait) {
            try {
                wait();
            }
            catch (InterruptedException e) {
            }
        }
    }

    public synchronized void turnTo(float heading, boolean wait) {
        targetHeading = normalizeAngle(heading);
        state = ROTATE;
        if (wait) {
            try {
                wait();
            }
            catch (InterruptedException e) {
            }
        }
    }

    public synchronized void go(float heading) {
        targetHeading = normalizeAngle(heading);
        state = GO;
    }

    public synchronized void stop() {
        leftMotor.setPower(Motor.STOP);
        rightMotor.setPower(Motor.STOP);
        state = STOP;
    }

    private float normalizeAngle(float angle) {
        while (angle < -PI)
            angle += TWO_PI;
        while (angle > PI)
            angle -= TWO_PI;
        return angle;
    }

    public void run() {
        try {
            while (true) {
                switch (state) {
                case MOVE_TO:
                    goToPoint();
                    break;
                case GO:
                    goHeading();
                    break;
                case ROTATE:
                    doRotate();
                    break;
                default: // stopped
                    break;
                }
                Thread.sleep(period);
            }
        }
        catch (Throwable t) {
            t.printStackTrace();
        }
    }

    private synchronized void goHeading() {
        Pose pose = localizer.getPose();
        float error = targetHeading - pose.heading;
        if (error > PI)
            error -= TWO_PI;
        else if (error < -PI)
            error += TWO_PI;

        int offset = (int) (gain * error + 0.5f);

        leftMotor.setPower(drivePower - offset);
        rightMotor.setPower(drivePower + offset);
    }

    private synchronized void goToPoint() {
        Pose pose = localizer.getPose();
        float xError = destinationX - pose.x;
        float yError = destinationY - pose.y;

        float absXError = (xError > 0.0f) ? xError : -xError;
        float absYError = (yError > 0.0f) ? yError : -yError;
        if ((absXError + absYError) < goToThreshold) {
            // stop
            leftMotor.setPower(Motor.STOP);
            rightMotor.setPower(Motor.STOP);
            state = STOP;

            // signal waiting thread we are at the destination
            notify();
        }
        else {
            // adjust heading and continue on
            targetHeading = (float) Math.atan2(yError, xError);
            goHeading();
        }
    }

    private synchronized void doRotate() {
        Pose pose = localizer.getPose();
        float error = targetHeading - pose.heading;
        // choose the direction of rotation that results
        // in the smallest angle
        if (error > PI)
            error -= TWO_PI;
        else if (error < -PI)
            error += TWO_PI;
        float absError = (error >= 0.0f) ? error : -error;
        if (absError < rotateThreshold) {
            // stop
            leftMotor.setPower(Motor.STOP);
            rightMotor.setPower(Motor.STOP);
            state = STOP;

            // signal waiting thread we are at the destination
            notify();
        }
        else if (error > 0.0f) {
            // rotate counterclockwise
            leftMotor.setPower(-rotatePower);
            rightMotor.setPower(rotatePower);
        }
        else {
            // rotate clockwise
            leftMotor.setPower(rotatePower);
            rightMotor.setPower(-rotatePower);
        }
    }
}
