/*
 * LineFollower2.java
 * 
 * Copyright 2007 by RidgeSoft, LLC., PO Box 482, Pleasanton, CA 94566, U.S.A.
 * www.ridgesoft.com
 * 
 * RidgeSoft grants you the right to use, modify, make derivative works and
 * redistribute this source file provided you do not remove this copyright
 * notice.
 */

import com.ridgesoft.intellibrain.IntelliBrain;
import com.ridgesoft.io.Display;
import com.ridgesoft.robotics.AnalogInput;
import com.ridgesoft.robotics.PushButton;
import com.ridgesoft.robotics.Motor;
import com.ridgesoft.robotics.ContinuousRotationServo;

public class LineFollower2 {
    private static final int THRESHOLD = 300;

    // States
    private static final byte LOST = 0;
    private static final byte BOTH_LEFT = 1;
    private static final byte ONE_LEFT = 2;
    private static final byte CENTERED = 3;
    private static final byte ONE_RIGHT = 4;
    private static final byte BOTH_RIGHT = 5;

    // State transition table
    //      0 - sensor is not over line
    //      1 - sensor is over the line
    private static final byte[][] NEXT_STATE = new byte[][] {
            //             Sensor Conditions Code (left, right)
            //          (low, low)  (low, hi) (hi, low)  (hi, hi)
            // binary    00          01        10         11             Current
            // decimal   0           1         2          3               State
            new byte[] { LOST,       ONE_LEFT, ONE_RIGHT, CENTERED }, // LOST
            new byte[] { BOTH_LEFT,  ONE_LEFT, ONE_RIGHT, CENTERED }, // BOTH_LEFT
            new byte[] { BOTH_LEFT,  ONE_LEFT, ONE_RIGHT, CENTERED }, // ONE_LEFT
            new byte[] { LOST,       ONE_LEFT, ONE_RIGHT, CENTERED }, // CENTERED
            new byte[] { BOTH_RIGHT, ONE_LEFT, ONE_RIGHT, CENTERED }, // ONE_RIGHT
            new byte[] { BOTH_RIGHT, ONE_LEFT, ONE_RIGHT, CENTERED }, // BOTH_RIGHT
    };

    // Motor power levels
    private static final byte NORMAL = 10;
    private static final byte LOW = 5;

    // Left and right array indicies
    private static final byte LEFT = 0;
    private static final byte RIGHT = 1;

    // Action table
    private static final byte[][] POWER = new byte[][] {
        //           Left Right           State
        new byte[] { 0,      0      }, // LOST
        new byte[] { NORMAL, 0      }, // BOTH_LEFT
        new byte[] { NORMAL, LOW    }, // ONE_LEFT
        new byte[] { NORMAL, NORMAL }, // CENTERED
        new byte[] { LOW,    NORMAL }, // ONE_RIGHT
        new byte[] { 0,      NORMAL }, // BOTH_RIGHT
    };

    public static void main(String args[]) {
        Display display = IntelliBrain.getLcdDisplay();

        AnalogInput leftLineSensor = IntelliBrain.getAnalogInput(6);
        AnalogInput rightLineSensor = IntelliBrain.getAnalogInput(7);

        Motor leftMotor = new ContinuousRotationServo(IntelliBrain.getServo(1), 
                    false, 14);
        Motor rightMotor = new ContinuousRotationServo(IntelliBrain.getServo(2), 
                    true, 14);

        PushButton startButton = IntelliBrain.getStartButton();
        startButton.waitReleased();

        int state = LOST;
        String stateName = "Starting";
        boolean go = false;

        while (true) {
            // Sample sensors
            int leftSample = leftLineSensor.sample();
            int rightSample = rightLineSensor.sample();

            // Convert sensor readings to conditions code
            //   Conditions (left, right)   (low, low)  (low, hi)  (hi, low)   (hi, hi)  
            //   Binary                             00         01         10         11
            //   Decimal                             0          1          2          3
            int conditions = 0;
            if (leftSample > THRESHOLD)
                conditions |= 0x2;
            if (rightSample > THRESHOLD)
                conditions |= 0x1;

            // Lookup the new state
            state = NEXT_STATE[state][conditions];

            if (go) {
                // Perform action according to state
                leftMotor.setPower(POWER[state][LEFT]);
                rightMotor.setPower(POWER[state][RIGHT]);
            }
            else if (startButton.isPressed()) {
                go = true;
            }

            // Display sensor readings
            display.print(0, Integer.toString(leftSample) + ' ' + Integer.toString(rightSample));

            // Display current state
            switch (state) {
            case BOTH_LEFT:
                stateName = "Both Left";
                break;
            case ONE_LEFT:
                stateName = "One Left";
                break;
            case CENTERED:
                stateName = "Centered";
                break;
            case ONE_RIGHT:
                stateName = "One Right";
                break;
            case BOTH_RIGHT:
                stateName = "Both Right";
                break;
            case LOST:
                stateName = "Lost";
                break;
            }
            display.print(1, stateName);
        }
    }
}